// ball.hh - most stuff having to do with balls
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef BALL_HH
#define BALL_HH

#include <list>
#include <string>
#include <typeinfo>

#include "controller.hh"
#include "graphics.hh"
#include "record.hh"
#include "misc.hh"

// needed for balls::the_ship
namespace Game { class game; }

namespace balls {

    using misc::vectx;
    using misc::vecty;
    using misc::vectz;
    using misc::vect2d;
    using misc::vect3d;
    using misc::square;
    using misc::random;
    
    class ball;

    // used to insert new ball*s, e.g., fragments, bullets
    class swallow_ball {
    public:
        virtual void swallow(ball* b) = 0;
        virtual ~swallow_ball() { }
    };

    // swallow ball* into an inserter iterator
    template <typename Ins>
    class swallow_ball_ins : public swallow_ball {
    public:
        swallow_ball_ins(Ins i) : ins(i) { }
        void swallow(ball* b) { *ins = b; }
    private:
        Ins ins;
    };

    // returns a swallow_ball*. After usage, this should be freed
    // using operator delete. This is a helper function to create a
    // swallow_ball_ins
    template <typename Ins>
    swallow_ball* make_swallower(Ins i)
    {
        return new swallow_ball_ins<Ins>(i);
    }

    // info needed by a ball
    struct ball_info {
        rec::recrecord* bd;
        swallow_ball* swal;
        misc::random* rand;
    };

    struct draw_ball_info : public ball_info {
        graphics::deep_image_map* gallery;
        graphics::image_map* wicons;
        const misc::mapper<>* mapp;
    };

    // Abstract base class ball
    //    Things common to all balls, such as radius, mass and
    //    swallow_ball*, which is used to insert new balls, e.g.,
    //    fragments from a broken ball.
    // pure virtual functions:
    //    move(...): move for a given time
    //    do_collide(...): take care of collision with another ball.
    //    damage(): the damage caused to other balls bumping into this.
    class ball {
    public:
        ball(int r, int m, vect3d<> ll, vect3d<> vv, ball_info* inf)
            : rem(false), rad(r), mas(m), l(ll), v(vv), info(inf)
        {
        }
   
        virtual ~ball()
        {
        }

        virtual void move(int t) = 0;
        virtual void do_collide(const ball* other) = 0;
        virtual int damage() const = 0;
   
        bool remove() const             { return rem; }
        double radius() const           { return rad; }
        int mass() const                { return mas; }
        vect3d<> loc() const            { return l; }
        vect3d<> vel() const            { return v; }
        ball_info* inf() const          { return info; }
   
        friend void bounce_wall(ball* b);
        friend void get_inside_wall(ball* b);

    protected:
        bool   rem;  // set when the ball is broken up
        double rad;
        int    mas;
        vect3d<> l;     // the centre point
        vect3d<> v;     // velocity after last move
        ball_info* info;
    };

    // helper functions

    // check if this ball hits the tunnel wall (tunnel radius == tunnel_r)
    inline bool hit_wall(const ball* b)
    {
        return square(b -> loc().x()) + square(b -> loc().y())
            > square(tunnel_r - b -> radius());
    }

    // inline functions

    // check if top of b1 is above top ofb2
    inline bool above(const ball* b1, const ball* b2)
    {
        return b1->loc().z() - b1->radius() < b2->loc().z() - b2->radius();
    }
   
    // if top of b1 is below bottom of b2, i.e., b1 completely below b2
    inline bool well_below(const ball* b1, const ball* b2)
    {
        return b1->loc().z() - b1->radius() > b2->loc().z() + b2->radius();
    }

    // friends
    void bounce_wall(ball* b);
    void get_inside_wall(ball* b);

    // others
    bool collide(const ball* b1, const ball* b2);
    vect3d<> dvel_collision(const ball* b1, const ball* b2);
    
    // Now come the balls

    // A ship, not an enemy. This still is not final, and it cannot be
    // drawn. This class provides basic functionality needed by all
    // ships.
    class weapon;
    class ship : public ball {
    public:
        friend class weapon;
        ship(double r, int m, vect3d<> ll, vect3d<> vv, ball_info* inf,
             int str, int hrd, int amm)
            : ball(r, m, ll, vv, inf), score(0),
              strength(str), hardness(hrd), ammo(amm), dv(0, 0, 0)
        { }

        void add_score(int i) { score += i;}

        void move(int t);
        void do_collide(const ball* other);
        int damage() const;

    protected:
        int score;
        int strength, hardness, ammo;
        vect3d<> dv;
    };

    inline bool is_ship(const ball* b)
    {
        return dynamic_cast<const ship*>(b) != 0;
    }

    // abstract base class to hold information about weapon
    class weapon {
    public:
        weapon(ship* sp) : s(sp) { }
        virtual void charge(int t) = 0;
        virtual void fire() = 0;
        virtual const graphics::image* get_image() const = 0;
    protected:
        int& ammo() { return s->ammo; }
        ship* s;
    };

    // base class for bullets. Note that constructor takes location
    // and velocity.
    class bullet : public ball {
    public:
        bullet(double r, int m, vect3d<> ll, vect3d<>vv, ball_info* inf,
               int d)
            : ball(r, m, ll, vv, inf), hit(false), dam(d)
            { l = ll; v = vv; }

        void move(int t);
        void do_collide(const ball* other);
        int damage() const;
    protected:
        bool hit;
        int dam;
    };

    inline bool is_bullet(const ball* b)
    {
        return dynamic_cast<const bullet*>(b) != 0;
    }

    class ship_bullet : public bullet {
    public:
        ship_bullet(double r, int m, vect3d<> ll, vect3d<> vv, ball_info* inf,
                    int d, ship* s)
            : bullet(r, m, ll, vv, inf, d), firer(s)
        {
        }

        virtual void add_score(int i) const { firer->add_score(i); }
    protected:
        ship* firer;
    };

    inline bool is_ship_bullet(const ball* b)
    {
        return dynamic_cast<const ship_bullet*>(b) != 0;
    }

    // base class for pickups. 
    class pickup : public ball {
    public:
        pickup(double r, int m, vect3d<> ll, vect3d<>vv, ball_info* inf, int c)
            : ball(r, m, ll, vv, inf), cont(c), dv(0, 0, 0)
        { }
   
        void move(int t);
        void do_collide(const ball* other);
        int damage() const;

        virtual int pick(int request);
    protected:
        int cont;
        vect3d<> dv;
    };

    inline bool is_pickup(const ball* b)
    {
        return dynamic_cast<const pickup*>(b) != 0;
    }

    // base for other enemy balls.
    class enemy : public ball {
    public:
        enemy(double r, int m, vect3d<> ll, vect3d<>vv, ball_info* inf,
              int str, int hrd, int scr)
            : ball(r, m, ll, vv, inf),
              strength(str), hardness(hrd), score(scr), dv(0, 0, 0)
        { }

        void move(int t);
        void do_collide(const ball* other);
        int damage() const;

    protected:
        int strength, hardness, score;
        vect3d<> dv;
    };

    inline bool is_enemy(const ball* b)
    {
        return dynamic_cast<const enemy*>(b) != 0;
    }

    // drawable balls are derived from graphics::deep_drawable
    inline bool is_drawable(const ball* b)
    {
        return dynamic_cast<const graphics::drawable*>(b) != 0;
    }

    // function object to draw a ball
    struct do_ball_draw {
        void operator()(const ball* b)
        {
            const graphics::drawable* d;
            if ((d = dynamic_cast<const graphics::drawable*>(b)) != 0)
                d -> draw(b->loc());
        }
    };

    // The actual ship implementation. Takes care of controller (such
    // as keyboard), etc. Derived from ship and deep_drawable.
    class the_ship : public ship, public graphics::deep_drawable {
        friend class Game::game;
    public:
        the_ship(vect3d<> ll, vect3d<> vv, draw_ball_info* inf,
                 control::controller* controller);
        ~the_ship();
   
        void move(int t);
        
    protected:
        double min_zvel, max_zvel;         // min and max z velocities
        double zaccel, zdecel;             // z acceleration/decelaration
        double xyvel;                      // velocity for x/y movement
        double depth_scr;                  // score for a depth of 1.0
        int depth_scr_till_now;            // score added for depth till now
        control::controller* control;      // controller for movement
        std::list<weapon*> weapons;       // available weapons
        std::list<weapon*>::iterator cur_weap;
        bool changed_weap_last_turn;
    };

    class def_weapon : public weapon {
    public:
        def_weapon(ship* sp);
        void charge(int t);
        void fire();
        const graphics::image* get_image() const { return im; }
    private:
        int ch;
        int req_charge;
        int req_ammo;
        double extra_vel;
        const graphics::image* im;
    };

    class fast_weapon : public weapon {
    public:
        fast_weapon(ship* sp);
        void charge(int t);
        void fire();
        const graphics::image* get_image() const { return im; }

    private:
        int ch;
        int req_charge;
        int req_ammo;
        double extra_vel;
        const graphics::image* im;
    };

    // class normal_ball
    // defines things in which a normal enemy ball differs from base
    // class enemy
    // static public members
    //    new_ball(...): creates a new normal_ball
    //    and various constants
    class normal_ball : public enemy, public graphics::deep_drawable {
    public:
        normal_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
        void move(int t);

        static ball* new_ball(vect3d<> ll, vect3d<> vv, ball_info* inf)
        { return new normal_ball(ll, vv, static_cast<draw_ball_info*>(inf)); }
    };

    class normal_frag : public bullet, public graphics::deep_drawable {
    public:
        normal_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
    };

    // hard_ball
    //    similar to normal_ball, but its constants give a harder ball
    class hard_ball : public enemy, public graphics::deep_drawable {
    public:
        hard_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
        void move(int t);

        static ball* new_ball(vect3d<> ll, vect3d<> vv, ball_info* inf)
        { return new hard_ball(ll, vv, static_cast<draw_ball_info*>(inf)); }
    };

    class hard_frag : public bullet, public graphics::deep_drawable {
    public:
        hard_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
    };

    // fast_ball
    //    similar to normal_ball, but its constants give a faster ball
    class fast_ball : public enemy, public graphics::deep_drawable {
    public:
        fast_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
        void move(int t);

        static ball* new_ball(vect3d<> ll, vect3d<> vv, ball_info* inf)
        { return new fast_ball(ll, vv, static_cast<draw_ball_info*>(inf)); }
    };

    class fast_frag : public bullet, public graphics::deep_drawable {
    public:
        fast_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
    };

    // big_ball
    //    similar to normal_ball, but its constants give a bigger and
    //    harder ball
    class big_ball : public enemy, public graphics::deep_drawable {
    public:
        big_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
        void move(int t);

        static ball* new_ball(vect3d<> ll, vect3d<> vv, ball_info* inf)
        { return new big_ball(ll, vv, static_cast<draw_ball_info*>(inf)); }
    };

    class big_frag : public bullet, public graphics::deep_drawable {
    public:
        big_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
    };

    // whiz_ball
    //    similar to normal_ball, but its constants give a crazy ball
    class whiz_ball : public enemy, public graphics::deep_drawable {
    public:
        whiz_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
        void move(int t);

        static ball* new_ball(vect3d<> ll, vect3d<> vv, ball_info* inf)
        { return new whiz_ball(ll, vv, static_cast<draw_ball_info*>(inf)); }
    protected:
        int time_change, time_left;
    };

    class whiz_frag : public bullet, public graphics::deep_drawable {
    public:
        whiz_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf);
    };

    // class normal_bullet
    // defines things in which a normal bullet differs from base
    // class bullet
    // static public members
    //    various constants
    class normal_bullet : public ship_bullet, public graphics::deep_drawable {
    public:
        normal_bullet(vect3d<> ll, vect3d<> vv, draw_ball_info* inf, ship* s);
    };

    //////////////////////////////////////////////////////////////////////
    // The following section deals with creation of new balls
    //////////////////////////////////////////////////////////////////////

    // a function which gives a ball*
    typedef ball* (*new_ball_fn)(vect3d<> ll, vect3d<> vv, ball_info* inf);

    new_ball_fn name_to_nbf(const std::string& name);

    class creator {
    public:
        virtual ~creator() { }
        virtual void move(int t) = 0;
    };

} // namespace balls

#endif // !BALL_H
