// misc.hh - contains various miscellaneous classes
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MISC_HH
#define MISC_HH

#include <algorithm>
#include <cmath>
#include <iterator>
#include <numeric>
#include <typeinfo>
#include <vector>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

// First set the standards.
// One time unit is 1 ms.
// The tunnel has diameter of 1000 units and depth of 4000 units.
// Thus, with a velocity of 1, it takes 1 sec to move one
// tunnel diameter.

const int unit_msec   = 1;
const int unit_usec   = 1000;

const double tunnel_d    = 1000.0;
const double tunnel_dpth = 4000.0;
const double tunnel_r    = tunnel_d / 2;

namespace misc {

    // some functions:
    // square:    squares a number
    // diff:      returns the difference between two numbers
    // force_rng: forces a number within a given range

    // squares a number
    template <typename T>
    inline T square(const T& t)
    {
        return t * t;
    }
    
    // returns the difference between two numbers
    template <typename T>
    inline T diff(const T& t1, const T& t2)
    {
        return t1 > t2 ? t1 - t2 : t2 - t1;
    }

    // returns obj if it is within min and max, min if obj < min, 
    // or max if obj > max
    template <typename T>
    inline T force_rng(const T& obj, const T& min, const T& max)
    {
        return obj < min ? min : (obj > max ? max : obj);
    }

    // a vectx<> is a single number in the x direction which can be added to
    // 2d and 3d vectors
    template <typename T = double>
    class vectx {
    public:
        vectx(T xi = 0) : X(xi) { }
        
        template <typename U>
        vectx(const vectx<U>& v) : X(v.x()) { }

        vectx& set(T xi)  { X = xi; return *this; }
        
        T x() const     { return X; }
   
    private:
        double X;
    };

    // a vecty<> is similar to vectx<> but in the y direction
    template <typename T = double>
    class vecty {
    public:
        vecty(T yi = 0) : Y(yi) { }
        
        template <typename U>
        vecty(const vecty<U>& v) : Y(v.y()) { }

        vecty& set(T yi)  { Y = yi; return *this; }
        
        T y() const     { return Y; }
   
    private:
        double Y;
    };

    // a vectz<> is similar to vectx<> but in the z direction
    template <typename T = double>
    class vectz {
    public:
        vectz(T zi = 0) : Z(zi) { }
        
        template <typename U>
        vectz(const vectz<U>& v) : Z(v.z()) { }

        vectz& set(T zi)  { Z = zi; return *this; }
        
        T z() const     { return Z; }
   
    private:
        double Z;
    };

    // class vect2d
    //    two dimensional vector with x and y directions
    // helper functions
    //    mkpolar: makes a vector from polar coordinates
    //    square:  gives the square
    //    mag:     gives the magnitude
    //    arg:     gives the argument
    //    and several operators

    template <typename T = double>
    class vect2d {
    public:
        vect2d(T xi = 0, T yi = 0) : X(xi), Y(yi) { }
        vect2d(vectx<T> v) : X(v.x()), Y(0) { }
        vect2d(vecty<T> v) : X(0), Y(v.y()) { }

        template <typename U>
        vect2d(const vect2d<U>& v) : X(v.x()), Y(v.y()) { }

        vect2d& set(T xi, T yi)  { X = xi; Y = yi; return *this; }
        vect2d& set(const vect2d& v) { X = v.X; Y = v.Y; return *this; }
        vect2d& setx(T xi)       { X = xi; return *this; }
        vect2d& sety(T yi)       { Y = yi; return *this; }
        
        vect2d& operator=(const vect2d& v)
            { X = v.X; Y = v.Y; return *this; }
        vect2d& operator+=(const vect2d& v)
            { X += v.X; Y += v.Y; return *this; }
        vect2d& operator-=(const vect2d& v)
            { X -= v.X; Y -= v.Y; return *this; }
        vect2d& operator+=(const vectx<T>& v)
            { X += v.x(); return *this; }
        vect2d& operator+=(const vecty<T>& v)
            { Y += v.y(); return *this; }
        vect2d& operator-=(const vectx<T>& v)
            { X -= v.x(); return *this; }
        vect2d& operator-=(const vecty<T>& v)
            { Y -= v.y(); return *this; }
        template <typename U>
        vect2d& operator*=(const U& u)
            { X *= u; Y *= u; return *this; }
        template <typename U>
        vect2d& operator/=(const U& u)
            { X /= t; Y /= u; return *this; }

        T x() const     { return X; }
        T y() const     { return Y; }
   
        operator bool() const         { return X || Y; }
        bool operator!() const        { return !X && !Y; }
   
    private:
        double X, Y;
    };

    template <typename T>
    inline vect2d<T> mkpolar(const T& mag, const T& arg)
         { return vect2d<T>(mag * std::cos(arg), mag * std::sin(arg)); }

    template <typename T>
    inline T square(const vect2d<T>& v)
        { return square(v.x()) + square(v.y()); }
    
    template <typename T>
    inline T mag(const vect2d<T>& v)
        { return std::sqrt(square(v)); }

    template <typename T>
    inline T arg(const vect2d<T>& v)
        { return std::atan2(v.y(), v.x()); }

    template <typename T>
    inline vect2d<T> operator-(const vect2d<T>& v)
        { return vect2d<T>(-v.x(), -v.y()); }

    template <typename T>
    inline vect2d<T> operator+(const vect2d<T>& v)
        { return v; }

    template <typename T>
    inline bool operator==(const vect2d<T>& v1, const vect2d<T>& v2)
        { return v1.x() == v2.x() && v1.y() == v2.y(); }
    
    template <typename T>
    inline bool operator!=(const vect2d<T>& v1, const vect2d<T>& v2)
        { return v1.x() != v2.x() || v1.y() != v2.y(); }
    
    template <typename T, typename U>
    inline vect2d<T> operator+(const vect2d<T>& v, const U& u)
        { vect2d<T> r = v; r += u; return r; }
    
    template <typename T, typename U>
    inline vect2d<T> operator-(const vect2d<T>& v, const U& u)
        { vect2d<T> r = v; r -= u; return r; }
    
    template <typename T, typename U>
    inline vect2d<T> operator*(const vect2d<T>& v, const U& u)
        { vect2d<T> r = v; r *= u; return r; }
    
    template <typename T, typename U>
    inline vect2d<T> operator*(U u, const vect2d<T>& v)
        { vect2d<T> r = v; r *= u; return r; }
    
    template <typename T, typename U>
    inline vect2d<T> operator/(const vect2d<T>& v, const U& u)
        { vect2d<T> r = v; r /= u; return r; }
    
    //  class vect3d
    //     a three dimensional vector, with x, y and z
    //  helper functions
    //     mkpolar: sort of mkpolar for vect2d
    //     square:  gives the square
    //     mag:     gives the magnitude
    //     dotp:    gives the dot product of two vectors
    //     crossp:  gives the cross product of two vectors
    //     triplep: gives the triple product of three vectors
    //     and several operators

    template <typename T = double>
    class vect3d {
    public:
        vect3d(T xi = 0, T yi = 0, T zi = 0) : X(xi), Y(yi), Z(zi) { }
        vect3d(const vect2d<T>& v, T zi = 0) : X(v.x()), Y(v.y()), Z(zi) { }
        vect3d(vectz<T> v) : X(0), Y(0), Z(v.z()) { }

        template <typename U>
        vect3d(const vect3d<U>& v) : X(v.x()), Y(v.y()), Z(v.z()) { }

        vect3d& set(T xi = 0, T yi = 0, T zi = 0)
            { X = xi; Y = yi; Z = zi; return *this; }
        vect3d& set(const vect2d<T>& v, T zi = 0)
            { X = v.x(); Y = v.y(); Z = zi; return *this; }
        vect3d& set(const vect3d<T>& v)
            { X = v.X; Y = v.Y; Z = v.Z; return *this; }
        vect3d& setxy(const vect2d<T>& v)
            { X = v.x(); Y = v.y(); return *this; }
        vect3d& setx(T xi)       { X = xi; return *this; }
        vect3d& sety(T yi)       { Y = yi; return *this; }
        vect3d& setz(T zi)       { Z = zi; return *this; }
        
        vect3d& operator=(const vect3d& v)
            { X = v.X; Y = v.Y; Z = v.Z; return *this; }
        vect3d& operator+=(const vect3d& v)
            { X += v.X; Y += v.Y; Z += v.Z; return *this; }
        vect3d& operator-=(const vect3d& v)
            { X -= v.X; Y -= v.Y; Z -= v.Z; return *this; }
        vect3d& operator+=(const vect2d<T>& v)
            { X += v.x(); Y += v.y(); return *this; }
        vect3d& operator-=(const vect2d<T>& v)
            { X -= v.x(); Y -= v.y(); return *this; }
        vect3d& operator+=(const vectx<T>& v)
            { X += v.x(); return *this; }
        vect3d& operator-=(const vectx<T>& v)
            { X -= v.x(); return *this; }
        vect3d& operator+=(const vecty<T>& v)
            { Y += v.y(); return *this; }
        vect3d& operator-=(const vecty<T>& v)
            { Y -= v.y(); return *this; }
        vect3d& operator+=(const vectz<T>& v)
            { Z += v.z(); return *this; }
        vect3d& operator-=(const vectz<T>& v)
            { Z -= v.z(); return *this; }
        template <typename U>
        vect3d& operator*=(const U& u)
            { X *= u; Y *= u; Z *= u; return *this; }
        template <typename U>
        vect3d& operator/=(const U& u)
            { X /= u; Y /= u; Z /= u; return *this; }

        T x() const     { return X; }
        T y() const     { return Y; }
        T z() const     { return Z; }
        vect2d<T> xy() const { return vect2d<T>(X, Y); }
   
        operator bool() const         { return X || Y || Z; }
        bool operator!() const        { return !X && !Y && !Z; }

    private:
        double X, Y, Z;
    };

    // mag is magnitude, argxy give angle on xy plane (from -pi to +pi)
    // and argz give other angle (from -pi/2 to +pi/2)
    template <typename T>
    inline vect3d<T> mkpolar(const T& mag, const T& argxy, const T& argz)
        { return vect3d<T>(mkpolar(mag * std::cos(argz), argxy),
                           mag * std::sin(argz)); }

    template <typename T>
    inline T square(const vect3d<T>& v)
        { return square(v.x()) + square(v.y()) + square(v.z()); }
    
    template <typename T>
    inline T mag(const vect3d<T>& v)
        { return std::sqrt(square(v)); }

    template <typename T>
    inline vect3d<T> operator-(const vect3d<T>& v)
        { return vect3d<T>(-v.x(), -v.y(), -v.z()); }

    template <typename T>
    inline vect3d<T> operator+(const vect3d<T>& v)
        { return v; }

    template <typename T>
    inline bool operator==(const vect3d<T>& v1, const vect3d<T>& v2)
        { return v1.x() == v2.x() && v1.y() == v2.y() && v1.z() == v2.z(); }
    
    template <typename T>
    inline bool operator!=(const vect3d<T>& v1, const vect3d<T>& v2)
        { return v1.x() != v2.x() || v1.y() != v2.y() || v1.z() != v2.z(); }
    
    template <typename T, typename U>
    inline vect3d<T> operator+(const vect3d<T>& v, const U& u)
        { vect3d<T> r = v; r += u; return r; }
    
    template <typename T, typename U>
    inline vect3d<T> operator-(const vect3d<T>& v, const U& u)
        { vect3d<T> r = v; r -= u; return r; }
    
    template <typename T, typename U>
    inline vect3d<T> operator*(const vect3d<T>& v, const U& u)
        { vect3d<T> r = v; r *= u; return r; }
    
    template <typename T, typename U>
    inline vect3d<T> operator*(U u, const vect3d<T>& v)
        { vect3d<T> r = v; r *= u; return r; }
    
    template <typename T, typename U>
    inline vect3d<T> operator/(const vect3d<T>& v, const U& u)
        { vect3d<T> r = v; r /= u; return r; }

    template <typename T>
    inline T dotp(const vect3d<T>& v1, const vect3d<T>& v2)
        { return v1.x() * v2.x() + v1.y() * v2.y() + v1.z() * v2.z(); }

    template <typename T>
    inline vect3d<T> crossp(const vect3d<T>& v1, const vect3d<T>& v2)
        { return vect3d<T>(v1.y() * v2.z() - v1.z() * v2.y(),
                           v1.z() * v2.x() - v1.x() * v2.z(),
                           v1.x() * v2.y() - v1.y() * v2.x()); }

    template <typename T>
    inline T triplep(const vect3d<T>& v1, const vect3d<T>& v2,
                     const vect3d<T>& v3)
        { return dotp(v1, crossp(v2, v3)); }
    
    ///////////////////////////////////////////////////////
    // Random stuff

    class random {
        // uses rand_r(), which is weak as all the state of the random
        // number gnerator has to be kept in an unsigned long
    public:
        random(unsigned int s = 1) : seed(s) { }

        // gives a random integer in the range [0, RAND_MAX)
        int operator()();
        int max() const;
        
        // gives a random integer in the range [0, m)
        int operator()(int m)
            { return int(double(m) * (*this)() / (max() + 1.0)); }

        // gives a random double in the range [0, m)
        double operator()(double m)
            { return (double(m) * (*this)() / (max() + 1.0)); }
    private:
        unsigned int seed;
    };
    
    // number in the range [-max, max)
    template <typename Rand>
    inline int rndpm(Rand& r, int max)
    { return r(max + max) - max; }

    template <typename Rand>
    inline double rndpm(Rand& r, double max)
    { return r(max + max) - max; }

    // number in the range [min, max)
    template <typename Rand>
    inline int rndrng(Rand& r, int min, int max)
    { return r(max - min) + min; }

    template <typename Rand>
    inline double rndrng(Rand& r, double min, double max)
    { return r(max - min) + min; }
   
    // location that is equally probable on all pts in a circle of
    // radius rad, such that r is in the range [0, rad)
    template <typename Rand, typename Flt>
    inline vect2d<Flt> rndpolar(Rand& rnd, Flt rad)
    { return mkpolar<Flt>(rad * std::sqrt(rnd(1.0)), rnd(2.0 * M_PI)); }

    template <typename Rand, typename Flt>
    inline vect3d<Flt> rnd3d(Rand& rnd, Flt rad)
    { return mkpolar<Flt>(rad * std::pow(rnd(1.0), 1.0/3),
                          rnd(M_PI), rnd(M_PI)); }

    // class rand_int
    //    used to give a random number from 0 to max - 1. The probabilities
    //    for the different numbers are given in the constructor or to the
    //    reset(...) function.
    // functions
    //    rand_int(...):   takes the start and end input iterators.
    //                     The size of the given range is the number of
    //                     possible outcomes, with each outcome having the
    //                     chance given by the number pointed to by the
    //                     iterator.
    //    reset(...):      does the same thing as the constructor
    //    operator()(...): gives a random int as described above using the
    //                     random generator given
    //    size():          gives the number of possible outcomes
    //    operator[](...): gives the chance of a specific outcome
    //    swap(...):       swap with another rand_int
    // NOTE
    //    The chances given in the constructor or reset(...), when added
    //    together, give a probability of 1. So if three numbers are given,
    //    20, 5 and 25, (total = 50), the return of operator()(...) has the
    //    following distribution:
    //       P(0) = 20/50 = 0.4
    //       P(1) =  5/50 = 0.1
    //       P(2) = 25/50 = 0.5
    class rand_int {
    public:
        rand_int()
            : total(0) { }
        rand_int(const rand_int& r)
            : limits(r.limits), total(r.total) { }

        template <typename In>
        rand_int(In start, In finish);

        template <typename In>
        void reset(In start, In finish);

        template <typename Rand>
        int operator()(Rand& r) const;
        
        size_t size() { return limits.size(); }

        int operator[](size_t n) const
        { if (n==0) return limits[0]; else return limits[n] - limits[n-1]; }
   
        void swap(rand_int& other)
        { limits.swap(other.limits); std::swap(total, other.total); }
        
    private:
        std::vector<int> limits;
        int total;
    };

    // Projection and mapping

    // projects distances according to the depth. This also converts
    // from a double to an int, as coords of objects are held as
    // double and they have to be drawn as int.
    template <typename Int = int, typename Flt = double>
    class projector {
    public:
        // If an object is at distance dist and has width ws and height hs,
        // it will be mapped to width wt and height ht.
        // So   projx = x * (wt / ws) * (dist / (dist + depth))
        //            = x * multx / (dist + depth), multx = wt * dist / ws
        // and  projy = y * multy / (dist + depth), multy = ht * dist / hs
        projector(Flt ws, Flt hs, Flt dist, Int wt, Int ht)
            : multx(wt * dist / ws), multy(ht * dist / hs), multb(0.1 / dist),
              distance(dist)
        { }

        Int projx(Flt x, Flt depth) const
        { return int(x * multx / (distance + depth)); }
   
        Int projy(Flt y, Flt depth) const
        { return int(y * multy / (distance + depth)); }

        // pass through points (0, 1), (a, b), (1, 0)
        // So for f = [0, a], r = 1 - f * (1-b)/a
        // and for f = [a, 1], r = (1-f) * b/(1-a)
        Flt projcol(Flt depth) const
        { Flt f = multb * depth;
          const double a = 0.8, b = 0.5;
          if (f < a) return 1 - f * (1-b)/a;
          if (f < 1) return (1-f) * b/(1-a);
          return 0; }

    private:
        Flt multx, multy, multb, distance;
    };
    
    // performs the job of a projector and gives a translation as well
    template <typename Int = int, typename Flt = double>
    class mapper : public projector<Int, Flt> {
    public:
        mapper(const projector<Int, Flt> &p, Int cx, Int cy)
            : projector<Int, Flt>(p),
              centrex(cx), centrey(cy),
              zero_depth(0)
        { }

        Int mapx(Flt x, Flt depth) const
        { return centrex + projx(x, depth - zero_depth); }

        Int mapy(Flt y, Flt depth) const
        { return centrey + projy(y, depth - zero_depth); }

        void set_depth(Flt depth)
        { zero_depth = depth; }
        Flt get_depth() const
        { return zero_depth; }
        
    private:
        Int centrex, centrey;
        Flt zero_depth;
    };

    // used to scale an image from a size of srcw * srch to a size of
    // dstw * dsth such that getx(dstw) returns srcw. In this way, the
    // value of the pixel at location (x, y) in the destination can be
    // assigned to the value of the pixel at location (getx(x), gety(y))
    // in the source image. The mapping is performed using a lookup table.
    class scaler1 {
    public:
        scaler1(int src, int dst) { set(src, dst); }
        ~scaler1() { clear(); }
        void reset(int src, int dst) { clear(); set(src, dst); }

        int get(int i) const { return map[i]; }

    private:
        int *map;

        void set(int src, int dst);
        void clear();
    };

    class scaler {
    public:
        scaler(int srcw, int srch, int dstw, int dsth)
            { set(srcw, srch, dstw, dsth); }
        ~scaler()
            { clear(); }
        void reset(int srcw, int srch, int dstw, int dsth)
            { clear(); set(srcw, srch, dstw, dsth); }

        // from a location in dst, give required location in src
        int getx(int x) const { return mapx[x]; }
        int gety(int y) const { return mapy[y]; }

        // for a size in src, give required size of dst
        int getw(int w) const { return w * dw / sw; }
        int geth(int h) const { return h * dh / sh; }

    private:
        int sw, sh, dw, dh;
        int *mapx, *mapy;

        void set(int srcw, int srch, int dstw, int dsth);
        void clear();
    };

    // function object to compare type_info pointers
    struct tip_cmp {
        bool operator()(const std::type_info* t1,
                        const std::type_info* t2) const
        { if (t2 == 0) return false; if (t1 == 0) return true;
          return t1 -> before(*t2); }
    };

} // namespace misc

// Store the cumulative sum of chances in the vector limits. Assign total
// to the total chance.
template <typename In>
misc::rand_int::rand_int(In start, In finish)
    : total(0)
{
    std::partial_sum(start, finish, std::back_inserter(limits));
    if (!limits.empty())
        total = limits.back();
}

// Similar to the constructor above.
template <typename In>
void misc::rand_int::reset(In start, In finish)
{
    total = 0;
    limits.clear();
    std::partial_sum(start, finish, std::back_inserter(limits));
    if (!limits.empty())
        total = limits.back();
}

// Get a random number according to limits.
template<typename Rand>
inline int misc::rand_int::operator()(Rand& rnd) const
{
    // r is a random number up to total
    int r = rnd(total);

    // i starts as 0 and goes up until it is the required intege
    int i = 0;
    for (std::vector<int>::const_iterator l = limits.begin();
         l != limits.end(); ++l)
        if (r < *l)
            return i;
        else
            ++i;
    return i;    // last line should never be met
}

#endif // MISC_HH
