// record.hh - storing of data
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef RECORD_HH
#define RECORD_HH

#include <algorithm>
#include <functional>
#include <map>
#include <string>
#include <typeinfo>
#include <exception>
#include <iostream>
#include <strstream>

#include "except.hh"

namespace rec {

    // conversions from string to bool, int and double and vice versa
    inline bool s2b(const std::string& s)
    {
        bool b;
        std::istrstream is(s.c_str());
        is >> b;
        return b;
    }

    inline int s2i(const std::string& s)
    {
        int i;
        std::istrstream is(s.c_str());
        is >> i;
        return i;
    }

    inline double s2d(const std::string& s)
    {
        double d;
        std::istrstream is(s.c_str());
        is >> d;
        return d;
    }

    inline std::string b2s(bool b)
    {
        char s[64]; // over large
        std::ostrstream os(s, 64);
        os << b << std::ends;
        return s;
    }

    inline std::string i2s(int i)
    {
        char s[64]; // over large
        std::ostrstream os(s, 64);
        os << i << std::ends;
        return s;
    }

    inline std::string d2s(double d)
    {
        char s[64]; // over large
        std::ostrstream os(s, 64);
        os << d << std::ends;
        return s;
    }

    class not_found : public std::exception { };
    class bad_type : public std::exception { };
    
    // abstract base class of holdert<>
    class holder {
    public:
        virtual ~holder() { }

        virtual operator bool() const = 0;
        virtual operator int() const = 0;
        virtual operator double() const = 0;
        virtual operator std::string() const = 0;

        virtual holder* copy() const = 0;
    };

    // holds a member of type T and can convert it to bool, int, double
    // and string
    template <typename T>
    class holdert : public holder {
    public:
        holdert() : t() { }
        holdert(const T& c) : t(c) { }
        holdert(const holdert& other) : t(other.t) { }

        holdert& operator=(const holdert& other) { t = other.t; }

        T& ref() { return t; }
        const T& ref() const { return t; }

        operator bool() const;
        operator int() const;
        operator double() const;
        operator std::string() const;

        holder* copy() const { return new holdert(t); }
    private:
        T t;
    };

    template <typename T>
    class tag {
    public:
        typedef T value_type;
        typedef T* pointer_type;
    };
    
    // holds an object of any type. Assignment gives the item the type
    // of the rvalue.
    class item {
    public:
        item() : ptr(new holdert<none>) { }
        item(const item& other) : ptr(other.ptr->copy()) { }

        template <typename T>
        item(const T& t) { ptr = new holdert<T>(t); }

        ~item() { delete ptr; }
        
        operator bool() const { return ptr->operator bool(); }
        operator int() const { return ptr->operator int(); }
        operator double() const { return ptr->operator double(); }
        operator std::string() const { return ptr->operator std::string(); }

        template <typename Tag>
        typename Tag::value_type& ref(Tag)
        { typedef typename Tag::value_type T;
          holdert<T>* tp = dynamic_cast<holdert<T>*>(ptr);
          if (tp == 0) throw bad_type();
          return tp->ref(); }

        template <typename Tag>
        const typename Tag::value_type& ref(Tag) const
        { typedef typename Tag::value_type T;
          holdert<T>* tp = dynamic_cast<holdert<T>*>(ptr);
          if (tp == 0) throw bad_type();
          return tp->ref(); }

        template <typename T>
        item& operator=(const T& t)
        { delete ptr; ptr = new holdert<T>(t); return *this; }

        item& operator=(const item& other)
        { if (&other != this) { delete ptr; ptr = other.ptr->copy(); }
          return *this; }

        template <typename Tag>
        bool is_type(Tag) const
        { return dynamic_cast<holdert<typename Tag::value_type>*>(ptr) != 0; }
    private:
        class none { };
        holder* ptr;
    };

    std::istream& operator>>(std::istream& is, item& i);
    std::ostream& operator<<(std::ostream& os, const item& i);

    // default conversions for all holdert<> are to throw an exception
    template <typename T>
    inline holdert<T>::operator bool() const { throw bad_type(); }
    template <typename T>
    inline holdert<T>::operator int() const { throw bad_type(); }
    template <typename T>
    inline holdert<T>::operator double() const { throw bad_type(); }
    template <typename T>
    inline holdert<T>::operator std::string() const { throw bad_type(); }

    // conversions of holdert from bool, int, double and string to each other
    template <>
    inline rec::holdert<bool>::operator bool() const { return t; }
    template <>
    inline rec::holdert<bool>::operator int() const { return t; }
    template <>
    inline rec::holdert<bool>::operator double() const { return t; }
    template <>
    inline rec::holdert<bool>::operator std::string() const { return b2s(t); }

    template <>
    inline rec::holdert<int>::operator bool() const { return bool(t); }
    template <>
    inline rec::holdert<int>::operator int() const { return t; }
    template <>
    inline rec::holdert<int>::operator double() const { return t; }
    template <>
    inline rec::holdert<int>::operator std::string() const { return i2s(t); }

    template <>
    inline rec::holdert<double>::operator bool() const { return bool(t); }
    template <>
    inline rec::holdert<double>::operator int() const { return int(t); }
    template <>
    inline rec::holdert<double>::operator double() const { return t; }
    template <>
    inline rec::holdert<double>::operator std::string() const
    { return d2s(t); }

    template <>
    inline rec::holdert<std::string>::operator bool() const { return s2b(t); }
    template <>
    inline rec::holdert<std::string>::operator int() const { return s2i(t); }
    template <>
    inline rec::holdert<std::string>::operator double() const
    { return s2d(t); }
    template <>
    inline rec::holdert<std::string>::operator std::string() const
    { return t; }

    template <typename Key, typename T, class Cmp = std::less<Key> >
    class basic_record : public std::map<Key, T, Cmp> {
    public:
        T& operator[](const Key& key)
        { return std::map<Key, T, Cmp>::operator[](key); }

        const T& operator[](const Key& key) const
        { const_iterator ci = find(key);
          if (ci == end()) throw not_found();
          return ci->second; }
    };

    typedef basic_record<std::string, item> record;
    typedef basic_record<std::string, record> recrecord;
    
    template <typename Key, typename T, typename Cmp>
    std::istream& operator>>(std::istream& is, basic_record<Key, T, Cmp>& r);
    template <typename Key, typename T, typename Cmp>
    std::ostream& operator<<(std::ostream& os,
                             const basic_record<Key, T, Cmp>& r);
 //namespace rec

 template <typename Key, typename T, typename Cmp>
 std::istream& rec::operator>>(std::istream& is, basic_record<Key, T, Cmp>& r)

{
    char c;
    
    try {
        if (!(is >> c) || c != '(')
            throw fail();

        for (bool ready = false; !ready; ) {
            if (!(is >> c))
                throw fail();
            if (c == ')')
                ready = true;
            else {
                Key key;
                T t;
                if (c != '(' || !(is >> key) || !(is >> t))
                    throw fail();

                r[key] = t;
                
                if (!(is >> c) || c != ')')
                    throw fail();
            }
        }
    }
    catch (fail) {
        is.setstate(std::ios::failbit);
        return is;
    }
    return is;
}

template <typename Key, typename T, typename Cmp>
std::ostream& rec::operator<<(std::ostream& os, const basic_record<Key, T, Cmp>& r)
{
    os << "(\n";
    for (basic_record<Key, T, Cmp>::const_iterator i = r.begin();
         i != r.end(); ++i) {
        os << "( " << i->first << " " << i->second << " )\n";
    }
    os << ")\n";
    return os;
}
}
#endif // !RECORD_HH
