// sdl_graphics.cc - functions from graphics using SDL
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <algorithm>
#include <utility>
#include <fstream>
#include <assert.h>
#include "SDL.h"
#include "SDL_image.h"

#include "misc.hh"
#include "sdl_graphics.hh"

// Reads an SDL_Surface from a file, and leaves it in its format.
// The first pixel is taken as the colorkey.
sdl::surface::surface(const std::string& filename, bool masked)
    : ref_count(new int(1))
{
    // TODO masked, colorkey, alpha...
    s = IMG_Load(filename.c_str());
    Uint32 col;
    getpix(static_cast<Uint8*>(s->pixels), col, s->format->BytesPerPixel);
    if (masked)
        SDL_SetColorKey(s, SDL_SRCCOLORKEY, col);
    else
        SDL_SetColorKey(s, 0, 0);
}

// Create a surface with the given pixel format from src. It is scaled
// such that it has width w and height h, and its brightness is
// amplified by bright.
// Also initializes a reference count to avoid copying, since once image
// is created, it cannot be modified.
sdl::surface::surface(const SDL_Surface* src, const SDL_PixelFormat* format,
                      int w, int h, double bright)
    : ref_count(new int(1))
{
    int srcbypp = src->format->BytesPerPixel;
    int srcw = src->w, srch = src->h;
    const Uint8* srcp = static_cast<Uint8*>(src->pixels);
    bool colkey = src->flags & SDL_SRCCOLORKEY;

    s = SDL_CreateRGBSurface(0, w, h, format->BitsPerPixel,
                             format->Rmask, format->Gmask,
                             format->Bmask, format->Amask);
    if (w == 0 || h == 0)
        return;
    int dstbypp = s->format->BytesPerPixel;
    Uint8* dstp = static_cast<Uint8*>(s->pixels);

    // create buffer for resizing lines
    int* buffer = new int[4 * w * srch];
    for (int i = 0; i < 4 * w * srch; ++i)
        buffer[i] = 0;

    Uint32 srccolkey = 0, dstcolkey = 0;
    if (colkey) {
        srccolkey = src->format->colorkey;
        Uint8 r, g, b;
        SDL_GetRGB(srccolkey, src->format, &r, &g, &b);
        dstcolkey = SDL_MapRGB(s->format, r, g, b);
    }

    int* bufferp = buffer;
    Uint8* onecolbuf = new Uint8[4 * srch];
    const Uint8** srccp = new const Uint8*[srch];
    srccp[0] = srcp;
    for (int i = 1; i < srch; ++i)
        srccp[i] = srccp[i - 1] + src->pitch;
    
    int wt;
    
    int index = srcw;
    int lower = 0, rem = 0;
    int nlower = srcw / w, nrem = srcw % w;

    bool read = false;

    for (int i = 0; i < srcw; ) {
        if (!read) {
            read = true;
            for (int j = 0; j < srch; ++j) {
                Uint32 col;
                getpixm(srccp[j], col, srcbypp);
                if (colkey && col == srccolkey) {
                    onecolbuf[j*4+3] = 1;
                }
                else {
                    onecolbuf[j*4+3] = 0;
                    SDL_GetRGB(col, src->format,
                               &onecolbuf[j*4],
                               &onecolbuf[j*4+1],
                               &onecolbuf[j*4+2]);
                }
            }
        }
        
        if (i == nlower)
            wt = nrem;
        else
            wt = w;
        if (i == lower)
            wt -= rem;

        int* bufferpc = bufferp;
        for (int j = 0; j < srch; ++j) {
            if (!onecolbuf[j*4+3]) {
                *(bufferpc++) += onecolbuf[j*4] * wt;
                *(bufferpc++) += onecolbuf[j*4+1] * wt;
                *bufferpc += onecolbuf[j*4+2] * wt;
                bufferpc += w * 4 - 2;
            }
            else {
                bufferpc[3] += wt;
                bufferpc += w * 4;
            }
        }

        if (i == nlower || (i == nlower-1 && nrem == 0)) {
            bufferp += 4;

            if (i != nlower) {
                ++i;
                read = false;
            }

            index += srcw;
            lower = nlower;
            rem = nrem;
            nlower = index / w;
            nrem = index % w;
        }
        else if (i != nlower) {
            ++i;
            read = false;
        }
    }

    delete[] srccp;
    delete[] onecolbuf;
    
    bufferp = buffer;
    int* rowbuffer = new int[4 * w];
    for (int i = 0; i < 4 * w; ++i)
        rowbuffer[i] = 0;

    index = srch;
    lower = rem = 0;
    nlower = srch / h; nrem = srch % h;

    Uint8* dstcp = dstp;
    for (int i = 0; i < srch; ) {
        if (i == nlower)
            wt = nrem;
        else
            wt = h;
        if (i == lower)
            wt -= rem;

        int* rowbufferc = rowbuffer;
        int* bufferpc = bufferp;
        for (int j = 0; j < w; ++j) {
            *(rowbufferc++) += *(bufferpc++) * wt;
            *(rowbufferc++) += *(bufferpc++) * wt;
            *(rowbufferc++) += *(bufferpc++) * wt;
            *(rowbufferc++) += *(bufferpc++) * wt;
        }

        if (i == nlower || (i == nlower-1 && nrem == 0)) {
            int* rowbufferc = rowbuffer;
            Uint8* dstcpc = dstcp;
            for (int j = 0; j < w; ++j) {
                int r = *rowbufferc;
                *(rowbufferc++) = 0;
                int g = *rowbufferc;
                *(rowbufferc++) = 0;
                int b = *rowbufferc;
                *(rowbufferc++) = 0;
                int a = *rowbufferc;
                *(rowbufferc++) = 0;
                int div = srcw*srch;
                if (a * 2 > div) { // a/div > 1/2 
                    setpixm(dstcpc, dstcolkey, dstbypp);
                }
                else {
                    div -= a;
                    Uint32 col = 0;
                    if (div)
                        col = SDL_MapRGB(s->format,
                                                Uint8(bright * r / div),
                                                Uint8(bright * g / div),
                                                Uint8(bright * b / div));
                    setpixm(dstcpc, col, dstbypp);
                }
            }
            dstcp += s->pitch;
            
            if (i != nlower) {
                ++i;
                bufferp += 4 * w;
            }
            
            index += srch;
            lower = nlower;
            rem = nrem;
            nlower = index / h;
            nrem = index % h;
        }
        else if (i != nlower) {
            ++i;
            bufferp += 4 * w;
        }
    }

    delete[] rowbuffer;
    delete[] buffer;

    // set the colorkey
    if (colkey)
        SDL_SetColorKey(s, SDL_SRCCOLORKEY | SDL_RLEACCEL, dstcolkey);
}

// Copy constructor increments reference count.
sdl::surface::surface(const surface& i)
    : ref_count(i.ref_count), s(i.s)
{
    ++ *ref_count;
}

// Copies an image and increments reference count accordingly.
sdl::surface& sdl::surface::operator=(const surface& i)
{
    // make sure we are not assigning an image to itself
    if (&i == this)
        return *this;

    // first decrement count and destroy if it becomes zero.
    -- *ref_count;
    if (*ref_count == 0)
        SDL_FreeSurface(s);

    // perform the copying and increment the reference count
    ref_count = i.ref_count;
    s = i.s;
    ++ *ref_count;

    return *this;
}

// decrease reference count and destroy if count becomes zero
sdl::surface::~surface()
{
    -- *ref_count;
    if (ref_count == 0)
        SDL_FreeSurface(s);
}

sdl::image::image(const surface* src, const SDL_PixelFormat* format,
                  int w, int h, double bright, tsurface* dst)
    : s(src->access(), format, w, h, bright),
      ts(dst)
{
}

// draw the image onto its target surface
void sdl::image::draw(int x, int y) const
{
    // src is the source of the blit and t is the target
    const SDL_Surface* src = s.access();
    SDL_Surface* t = ts->access();

    SDL_Rect srcrect;
    SDL_Rect dstrect;
    srcrect.x = srcrect.y = 0;
    srcrect.w = src->w;
    srcrect.h = src->h;
    dstrect.x = x;
    dstrect.y = y;

    // Now perform some checks. Maybe useless because balls should
    // always be in tunnel, but this is because we use
    // SDL_LowerBlit(), which has no checking, so it is safer to just
    // perform the checks.
    
    // First check for negative coords
    if (dstrect.x < 0) {
        if (-dstrect.x >= srcrect.w)
            return;
        srcrect.x -= dstrect.x;
        srcrect.w += dstrect.x;
        dstrect.x = 0;
    }
    if (dstrect.y < 0) {
        if (-dstrect.y >= srcrect.h)
            return;
        srcrect.y -= dstrect.y;
        srcrect.h += dstrect.y;
        dstrect.y = 0;
    }
    // Then check for width/height too large
    int excess = dstrect.x + srcrect.w - t->w; // excess width
    if (excess > 0) {
        if (excess >= srcrect.w)
            return;
        srcrect.w -= excess;
    }
    excess = dstrect.y + srcrect.h - t->h; // excess height
    if (excess > 0) {
        if (excess >= srcrect.h)
            return;
        srcrect.h -= excess;
    }

    dstrect.w = srcrect.w;
    dstrect.h = srcrect.h;

    // use LowerBlit because we have the correct rectangles, so no
    // need for checking, clipping, etc.
    SDL_LowerBlit(const_cast<SDL_Surface*>(src), &srcrect, t, &dstrect);
}

// draw the deep image onto its target surface
void sdl::deep_image::draw(const misc::vect3d<>& loc,
                           const misc::mapper<>& m) const
{
    double z = loc.z() - m.get_depth();
    // make sure depth is within visible range
    if (z < 0 || z >= tunnel_dpth)
        return;
    
    // src is the source of the blit and s is the target
    const SDL_Surface* src = geti(z).access();
    SDL_Surface* s = ts->access();

    SDL_Rect srcrect;
    SDL_Rect dstrect;
    srcrect.x = srcrect.y = 0;
    srcrect.w = src->w;
    srcrect.h = src->h;
    // use mapper m to translate loc into the location on the target
    dstrect.x = m.mapx(loc.x(), loc.z()) - src->w / 2;
    dstrect.y = m.mapy(loc.y(), loc.z()) - src->h / 2;

    // Now perform some checks. Maybe useless because balls should
    // always be in tunnel, but this is because we use
    // SDL_LowerBlit(), which has no checking, so it is safer to just
    // perform the checks.
    
    // First check for negative coords
    if (dstrect.x < 0) {
        if (-dstrect.x >= srcrect.w)
            return;
        srcrect.x -= dstrect.x;
        srcrect.w += dstrect.x;
        dstrect.x = 0;
    }
    if (dstrect.y < 0) {
        if (-dstrect.y >= srcrect.h)
            return;
        srcrect.y -= dstrect.y;
        srcrect.h += dstrect.y;
        dstrect.y = 0;
    }
    // Then check for width/height too large
    int excess = dstrect.x + srcrect.w - s->w; // excess width
    if (excess > 0) {
        if (excess >= srcrect.w)
            return;
        srcrect.w -= excess;
    }
    excess = dstrect.y + srcrect.h - s->h; // excess height
    if (excess > 0) {
        if (excess >= srcrect.h)
            return;
        srcrect.h -= excess;
    }

    dstrect.w = srcrect.w;
    dstrect.h = srcrect.h;

    // use LowerBlit because we have the correct rectangles, so no
    // need for checking, clipping, etc.
    SDL_LowerBlit(const_cast<SDL_Surface*>(src), &srcrect, s, &dstrect);
}

sdl::dial::dial(tsurface* t, const graphics::raw_image* src, int amin,
                int amax, int w, int h)
    : ts(t), bypp(t->access()->format->BytesPerPixel),
      pitch(t->access()->pitch),
      Amin(amin), range(amax - amin), W(w), H(h)
{
    image* i = static_cast<image*>(ts->make_image(src));
    const Uint8* p = static_cast<Uint8*>(i->access()->pixels);
    int bypp = i->access()->format->BytesPerPixel;
    getpixm(p, aminc, bypp);
    getpixm(p, currc, bypp);
    getpixm(p, maxc, bypp);
    getpixm(p, amaxc, bypp);
    delete i;
}

void sdl::dial::draw(int x, int y, int actual, int min, int max) const
{
    // check ranges
    if (min < Amin)
        min = Amin;
    if (actual < min)
        actual = min;
    if (max > range + Amin)
        max = range + Amin;
    if (actual > max)
        actual = max;
    
    Uint8* line = new Uint8[W * bypp];
    Uint8* ptr = line;
    
    // actual, min, max -= Amin, *= W/range
    // Round down min and round up actual, max
    min = (min - Amin) * W / range;
    actual = ((actual - Amin) * W + range - 1) / range;
    max = ((max - Amin) * W + range - 1) / range;

    setpixelsm(ptr, aminc, bypp, min);
    setpixelsm(ptr, currc, bypp, actual - min);
    setpixelsm(ptr, maxc, bypp, max - actual);
    setpixelsm(ptr, amaxc, bypp, W - max);

    SDL_Surface* s = ts->access();
    SDL_LockSurface(s);

    Uint8* dst = static_cast<Uint8*>(s->pixels) + s->pitch * y + bypp * x;
    for (int i = 0; i < H; ++i) {
        Uint8* linep = line;
        Uint8* dstp = dst;
        for (int j = 0; j < W*bypp/4; ++j) {
            *(dstp++) = *(linep++);
            *(dstp++) = *(linep++);
            *(dstp++) = *(linep++);
            *(dstp++) = *(linep++);
        }
        switch (W*bypp % 4) {
            case 3:
                *(dstp++) = *(linep++);
            case 2:
                *(dstp++) = *(linep++);
            case 1:
                *(dstp++) = *(linep++);
            case 0:
                ;
        }
        dst += pitch;
    }
    
    SDL_UnlockSurface(s);

    delete[] line;
}

// clear the surface with the given color
void sdl::tsurface::clear(int r, int g, int b)
{
    Uint32 col = SDL_MapRGB(s->format, r, g, b);
    SDL_Rect rect;
    rect.x = rect.y = 0;
    rect.w = w();
    rect.h = h();
    SDL_FillRect(s, &rect, col);
}

// Make an image from a file.
graphics::raw_image* sdl::tsurface::make_raw_image(const std::string& filename)
{
    return new surface(filename);
}

graphics::raw_image* sdl::tsurface::make_mask_raw_image(
                                               const std::string& filename)
{
    return new surface(filename, true);
}

// Make an image that renders onto this surface from a file.
graphics::image* sdl::tsurface::make_image(const graphics::raw_image* src)
{
    // create the image
    return new image(static_cast<const surface*>(src), s->format,
                     src->w(), src->h(), 1.0, this);
}

graphics::image* sdl::tsurface::make_image(const graphics::raw_image* src,
                                            const misc::scaler& sc)
{
    // create the image
    return new image(static_cast<const surface*>(src), s->format,
                     sc.getw(src->w()), sc.geth(src->h()), 1.0, this);
}

// Make a deep image that renders onto this surface from a file. The
// number of depths is given, as well as a projector and the width and
// height of the object. These (w and h) are projected using p.
graphics::deep_image* sdl::tsurface::make_deep_image
    (const graphics::raw_image* src, int n, const misc::projector<>& p,
     double w, double h)
{
    // First create three vectors containg the widths, heights and
    // brightness factors for all n images.
    std::vector<int> ww(n);
    std::vector<int> hh(n);
    std::vector<double> bb(n);
    for (int i = 0; i < n; ++i) {
        double depth = tunnel_dpth * i / n;
        ww[i] = p.projx(w, depth);
        hh[i] = p.projy(h, depth);
        bb[i] = p.projcol(depth);
    }

    // create the deep_image
    return new deep_image(static_cast<const surface*>(src),
                          s->format, n, ww.begin(),
                          hh.begin(), bb.begin(), this);
}

graphics::tunnel_drawable* sdl::tsurface::make_tunnel(const misc::mapper<>* m)
{
    return new tunnel_drawable(this, m, 32, 8);
}

graphics::dial* sdl::tsurface::make_dial(const graphics::raw_image* src,
                                         int amin, int amax, int w, int h)
{
    return new dial(this, src, amin, amax, w, h);
}

sdl::tunnel_drawable::tunnel_drawable(tsurface* t, const misc::mapper<>* m,
                                      int r, int s)
    : ts(t), mp(m), rings(r), sections(s), Rings(r * s)
{
    rx = m->projx(tunnel_r, 0);
    ry = m->projy(tunnel_r, 0);

    // worst case for one line - all in one colour
    int pitch = Rings + 2 + rx / 250;
    image = new unsigned char[pitch * ry];

    cols = new rgb[rings];
    brights = new double[rings];

    std::vector<graphics::line_ellipse> ellipses(Rings + 1);

    for (int i = 0; i < Rings + 1; ++i) {
        double depth = tunnel_dpth * i / Rings;
        int X = m->projx(tunnel_r, depth);
        int Y = m->projy(tunnel_r, depth);

        ellipses[i].reset(X, Y);
    }

    unsigned char* im = image;
    int inside1, inside2;

    for (int i = ry; i > 0; --i) {
        // first ellipse can always be nexted
        ellipses[0].next();

        // background
        write(im, rx - ellipses[0].x());
        inside1 = ellipses[0].x();

        // all rings
        for (int j = 1; j <= Rings; ++j) {
            if (i <= ellipses[j].y()) {
                inside2 = ellipses[j].x();
                ellipses[j].next();
                if (ellipses[j].x() == inside2)
                    --inside2;
            }
            else {
                inside2 = 0;
            }
            write(im, inside1 - inside2);
            inside1 = inside2;
        }

        // center
        write(im, inside1);
    }

    // assign cols and brights
    last_col = 0;
    first_col = 0;
    for (int i = 0; i < rings; ++i) {
        next_col();
        brights[i] = m->projcol(tunnel_dpth * i / rings);
    }

    first_col = 0;
    center_col = back_col = SDL_MapRGB(ts->access()->format, 0, 0, 0);

    required_depth = tunnel_dpth / Rings;
    current_depth = absolute_depth = 0;
    current_sections = sections;
}

sdl::tunnel_drawable::~tunnel_drawable()
{
    delete[] brights;
    delete[] cols;
    delete[] image;
}

void sdl::tunnel_drawable::draw(const misc::vect3d<>& loc) const
{
    SDL_Surface* s = ts->access();

    // prepare colors
    std::vector<Uint32> colors(rings);
    size_t col_index = first_col;
    for (int i = 0; i < rings; ++i) {
        rgb col = cols[col_index];
        if (++col_index == size_t(rings))
            col_index = 0;
        double bright = brights[i];
        col.r = Uint8(col.r * bright);
        col.g = Uint8(col.g * bright);
        col.b = Uint8(col.b * bright);
        colors[i] = SDL_MapRGB(s->format, col.r, col.g, col.b);
    }

    int cx = mp->mapx(loc.x(), loc.z());
    int cy = mp->mapy(loc.y(), loc.z());

    int pitch = s->pitch;
    int bypp = s->format->BytesPerPixel;

    SDL_LockSurface(s);

    // get top left corner
    Uint8* pyup = static_cast<Uint8*>(s->pixels)
        + (cy - ry) * pitch + (cx - rx) * bypp;
    // get bottom left corner
    Uint8* pydn = static_cast<Uint8*>(s->pixels)
        + (cy + ry - 1) * pitch + (cx - rx) * bypp;

    // start at beginning of image
    unsigned char* im = image;

    for (int i = 0; i < ry; ++i) {
        Uint8* pup = pyup;
        Uint8* pdn = pydn;

        // how much pixels do I draw?
        int n = read(im, 1);

        // draw background
        setpixelsm(pup, back_col, bypp, n);
        setpixelsm(pdn, back_col, bypp, n);

        // draw rings
        n = read(im, current_sections);
        setpixelsm(pup, colors[0], bypp, n);
        setpixelsm(pdn, colors[0], bypp, n);
        for (int j = 1; j < rings; ++j) {
            n = read(im, sections);
            if (n != 0) {
                setpixelsm(pup, colors[j], bypp, n);
                setpixelsm(pdn, colors[j], bypp, n);
            }
        }

        // draw center, and record position in im for right tunnel
        unsigned char* imr = im;
        n = read(im, 1 + sections - current_sections);
        if (n != 0) {
            n *= 2;
            setpixelsm(pup, center_col, bypp, n);
            setpixelsm(pdn, center_col, bypp, n);
        }
        
        // draw right part

        // draw rings
        for (int j = rings - 1; j > 0; --j) {
            n = readr(imr, sections);
            if (n != 0) {
                setpixelsm(pup, colors[j], bypp, n);
                setpixelsm(pdn, colors[j], bypp, n);
            }
        }
        n = readr(imr, current_sections);
        setpixelsm(pup, colors[0], bypp, n);
        setpixelsm(pdn, colors[0], bypp, n);

        // draw background
        n = readr(imr, 1);
        setpixelsm(pup, back_col, bypp, n);
        setpixelsm(pdn, back_col, bypp, n);

        // next line
        pyup += pitch;
        pydn -= pitch;
    }

    SDL_UnlockSurface(ts->access());
}

void sdl::tunnel_drawable::move(double d)
{
    current_depth += d - absolute_depth;
    absolute_depth = d;
    while (current_depth >= required_depth) {
        current_depth -= required_depth;
        if (--current_sections == 0) {
            current_sections = sections;
            next_col();
        }
    }
}

void sdl::tunnel_drawable::next_col()
{
    ++last_col;
    if (last_col == rings)
        last_col = 0;
    double d = double(last_col) * 3 / rings;
    rgb c;
    const int min = 50;
    const int max = 150;
    if (d < 1) {
        c.r = int((max - min) * d) + min;
        c.g = min;
        c.b = max + min - c.r;
    }
    else if (d < 2) {
        c.g = int((max - min) * d) + min * 2 - max;
        c.b = min;
        c.r = max + min - c.g;
    }
    else {
        c.b = int((max - min) * d) + min * 3 - max * 2;
        c.r = min;
        c.g = max + min - c.b;
    }
    cols[first_col] = c;
    if (++first_col == size_t(rings))
        first_col = 0;
}
