// sdl_graphics.hh - SDL graphics implementation
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef SDL_GRAPHICS_HH
#define SDL_GRAPHICS_HH

#include <cstddef>
#include "SDL.h"

#include "graphics.hh"

namespace sdl {

    // Set a pixel at *dst and move dst to next pixel
    inline void setpixm1(Uint8*& dst, Uint32 col)
    {
        *(dst++) = col;
    }
    
    inline void setpixm2(Uint8*& dst, Uint32 col)
    {
        * reinterpret_cast<Uint16*>(dst) = col;
        dst += 2;
    }
    
    inline void setpixm3(Uint8*& dst, Uint32 col)
    {
        Uint8* colp = reinterpret_cast<Uint8*>(&col);
        *(dst++) = *(colp++);
        *(dst++) = *(colp++);
        *(dst++) = *colp;
    }
    
    inline void setpixm4(Uint8*& dst, Uint32 col)
    {
        * reinterpret_cast<Uint32*>(dst) = col;
        dst += 4;
    }
    
    inline void setpixmn(Uint8*& dst, Uint32 col, int bypp)
    {
        memcpy(dst, &col, bypp);
        dst += bypp;
    }
    
    inline void setpixm(Uint8*& dst, Uint32 col, int bypp)
    {
        switch (bypp) {
            case 1:
                setpixm1(dst, col);
                break;
            case 2:
                setpixm2(dst, col);
                break;
            case 3:
                setpixm3(dst, col);
                break;
            case 4:
                setpixm4(dst, col);
                break;
            default:
                setpixmn(dst, col, bypp);
        }
    }
    
    // Get the pixel at *dst and move dst to next pixel
    inline void getpixm1(const Uint8*& dst, Uint32& col)
    {
        col = *(dst++);
    }
    
    inline void getpixm2(const Uint8*& dst, Uint32& col)
    {
        col = * reinterpret_cast<const Uint16*>(dst);
        dst += 2;
    }
    
    inline void getpixm3(const Uint8*& dst, Uint32& col)
    {
        col = 0;
        Uint8* colp = reinterpret_cast<Uint8*>(&col);
        *(colp++) = *(dst++);
        *(colp++) = *(dst++);
        *colp = *(dst++);
    }
    
    inline void getpixm4(const Uint8*& dst, Uint32& col)
    {
        col = * reinterpret_cast<const Uint32*>(dst);
        dst += 4;
    }
    
    inline void getpixmn(const Uint8*& dst, Uint32& col, int bypp)
    {
        col = 0;
        Uint8* colp = reinterpret_cast<Uint8*>(&col);
        for (; bypp > 0; --bypp)
            *(colp++) = *(dst++);
    }
    
    inline void getpixm(const Uint8*& dst, Uint32& col, int bypp)
    {
        switch (bypp) {
            case 1:
                getpixm1(dst, col);
                break;
            case 2:
                getpixm2(dst, col);
                break;
            case 3:
                getpixm3(dst, col);
                break;
            case 4:
                getpixm4(dst, col);
                break;
            default:
                getpixmn(dst, col, bypp);
        }
    }
    
    // Set pixels at *dst and move dst to next pixel
    inline void setpixelsm1(Uint8*& dst, Uint32 col, int n)
    {
        for (int i = n; i; --i)
            *(dst++) = col;
    }

    inline void setpixelsm2(Uint8*& dst, Uint32 col, int n)
    {
        Uint16* dstt = reinterpret_cast<Uint16*>(dst);
        for (int i = n/4; i; --i) {
            *(dstt++) = col;
            *(dstt++) = col;
            *(dstt++) = col;
            *(dstt++) = col;
        }
        switch (n%4) {
            case 3: *(dstt++) = col;
            case 2: *(dstt++) = col;
            case 1: *(dstt++) = col;
        }
        dst = reinterpret_cast<Uint8*>(dstt);
    }

    inline void setpixelsm3(Uint8*& dst, Uint32 col, int n)
    {
        Uint8* colp = reinterpret_cast<Uint8*>(&col);
        for (int i = n; i; --i) {
            *(dst++) = colp[0];
            *(dst++) = colp[1];
            *(dst++) = colp[2];
        }
    }

    inline void setpixelsm4(Uint8*& dst, Uint32 col, int n)
    {
        Uint32* dstt = reinterpret_cast<Uint32*>(dst);
        for (int i = n/4; i; --i) {
            *(dstt++) = col;
            *(dstt++) = col;
            *(dstt++) = col;
            *(dstt++) = col;
        }
        switch (n%4) {
            case 3: *(dstt++) = col;
            case 2: *(dstt++) = col;
            case 1: *(dstt++) = col;
        }
        dst = reinterpret_cast<Uint8*>(dstt);
    }

    inline void setpixelsmn(Uint8*& dst, Uint32 col, int bypp, int n)
    {
        for (int i = n; i; --i) {
            Uint8* colp = reinterpret_cast<Uint8*>(&col);
            for (int j = 0; j < bypp; ++j) {
                *(dst++) = *(colp++);
                *(dst++) = *(colp++);
                *(dst++) = *colp;
            }
        }
    }

    inline void setpixelsm(Uint8*& dst, Uint32 col, int bypp, int n)
    {
        switch (bypp) {
            case 1:
                setpixelsm1(dst, col, n);
                break;
            case 2:
                setpixelsm2(dst, col, n);
                break;
            case 3:
                setpixelsm3(dst, col, n);
                break;
            case 4:
                setpixelsm4(dst, col, n);
                break;
            default:
                setpixelsmn(dst, col, bypp, n);
        }
    }

    // Wrappers for above functions not to move dst.
    // (Only difference is that these take Uint8* instead of Uint8*&)
    inline void setpix(Uint8* dst, Uint32 col, int bypp)
        { setpixm(dst, col, bypp); }
    inline void getpix(const Uint8* dst, Uint32& col, int bypp)
        { getpixm(dst, col, bypp); }
    inline void setpixels(Uint8* dst, Uint32 col, int bypp, int n)
        { setpixelsm(dst, col, bypp, n); }
    
    // Make an image from a SDL_Surface. New size is w * h, and colours
    // are amplified by bright such that a bright value of 0 would
    // give a black image. The image is stored with the given display
    // format for fast blitting.
    class surface : public graphics::raw_image {
    public:
        surface(const std::string& filename, bool masked = false);
        surface(const SDL_Surface* src, const SDL_PixelFormat* format,
                int w, int h, double bright);
        surface(const surface& i);
        ~surface();

        surface& operator=(const surface& i);

        const SDL_Surface* access() const { return s; }

        int w() const { return s->w; }
        int h() const { return s->h; }
    private:
        int* ref_count;
        SDL_Surface* s;
    };

    class tsurface;
    
    // An implementation of Abstract Base Class (ABC) graphics::image.
    class image : public graphics::image {
    public:
        image(const surface* src, const SDL_PixelFormat* format,
              int w, int h, double bright, tsurface* dst);

        void draw(int x, int y) const;

        const SDL_Surface* access() const { return s.access(); }
        
    private:
        surface s;
        tsurface* ts;
    };

    // An implementation of Abstract Base Class (ABC) graphics::deep_image.
    // It has protected member function geti(...) to get the image
    // corresponding to a certain depth.
    class deep_image : public graphics::deep_image {
    public:
        template <typename In1, typename In2, typename In3>
        deep_image(const surface* src, const SDL_PixelFormat* format,
                   size_t n, In1 w, In2 h, In3 bright, tsurface* dst);

        void draw(const misc::vect3d<>& loc, const misc::mapper<>& m) const;
        
    protected:
        surface& geti(double depth)
            { return arr[size_t(size * depth / tunnel_dpth)]; }
        const surface& geti(double depth) const
            { return arr[size_t(size * depth / tunnel_dpth)]; }
    private:
        size_t size;
        std::vector<surface> arr;
        tsurface* ts;
    };

    class dial : public graphics::dial {
    public:
        dial(tsurface* t, const graphics::raw_image* src, int amin,
             int amax, int w, int h);

        void draw(int x, int y, int actual, int min, int max) const;
    private:
        tsurface* ts;
        int bypp, pitch;
        int Amin, range, W, H;
        Uint32 aminc, currc, maxc, amaxc;
    };

    // target surface, i.e., surface to draw on.
    // This is an implementation of ABC graphics::tsurface.
    // It also provides function access() for its SDL_Surface* to be known.
    class tsurface : public graphics::tsurface {
    public:
        tsurface(SDL_Surface* sur) : s(sur) { }

        void flip()   { SDL_Flip(s); }

        int w() const { return s->w; }
        int h() const { return s->h; }

        void clear(int r, int g, int b);

        graphics::raw_image* make_raw_image(const std::string& filename);
        graphics::raw_image* make_mask_raw_image(const std::string& filename);
        graphics::image* make_image(const graphics::raw_image* src);
        graphics::image* make_image(const graphics::raw_image* src,
                                    const misc::scaler& sc);
        graphics::deep_image* make_deep_image(const graphics::raw_image* src,
                    int n, const misc::projector<>& p, double w, double h);
        graphics::tunnel_drawable* make_tunnel(const misc::mapper<>* m);
        graphics::dial* make_dial(const graphics::raw_image* src, int amin,
                                  int amax, int w, int h);

        SDL_Surface* access() const { return s; }
    private:
        SDL_Surface* s;
    };

    // draw the tunnel
    class tunnel_drawable : public graphics::tunnel_drawable {
    public:
        tunnel_drawable(tsurface* t, const misc::mapper<>* m, int r, int s);
        ~tunnel_drawable();
        
        void draw(const misc::vect3d<>& loc) const;
        void move(double d);
    private:
        struct rgb { Uint8 r, g, b; };

        tsurface* ts;
        const misc::mapper<>* mp;

        int rings, sections, Rings, rx, ry;
        unsigned char* image;
        rgb* cols;
        double* brights;
        size_t first_col;
        Uint32 back_col;
        Uint32 center_col;

        double required_depth;
        double absolute_depth;
        double current_depth;
        int current_sections;

        static void write(unsigned char*& i, int n)
        { while (n >= 250) { *(i++) = 250; n -= 250; } *(i++) = n; }
        static int read(unsigned char*& i, int s)
        { int n = 0;
          while (s-- > 0) { n += *i; while (*(i++) == 250) n += *i; }
          return n; }
        static int readr(unsigned char*& i, int s)
        { int n = 0;
          while (s-- > 0) { n += *--i; while (*(i-1)==250) { n+=250; --i; } }
          return n; }

        int last_col;
        void next_col();
    };

} // namespace sdl

// constructor for sdl::deep_image. Takes an oimage& containing the
// source, the number of depths in the deep_image, the widths and heights
// of all the images, the amplification to brightness of all the images,
// and the tsurface on which to be drawn.
// n is the number of depths and w, h, bright are input iterators that
// are assumed to have n items each.
template <typename In1, typename In2, typename In3>
sdl::deep_image::deep_image(const surface* src, const SDL_PixelFormat* format,
                            size_t n, In1 w, In2 h, In3 bright, tsurface* dst)
    : size(n), ts(dst)
{
    // avoid reallocation
    arr.reserve(size);

    for (size_t i = 0; i < size; ++i) {
        // create images from the source.
        arr.push_back(surface(src->access(), format, *w, *h, *bright));
        ++w;
        ++h;
        ++bright;
    }
}

#endif // !SDL_GRAPHICS_HH
