// sdl_timer.cc - various functions for the timer implementation with SDL
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "SDL.h"

#include "sdl_timer.hh"

// Update the state of the timer, so it is up to date for other functions.
inline void sdl::timer::update()
{
    if (running) {
        Uint32 t = SDL_GetTicks();

        // Find the number of Ticks passed since last_time.
        // In the rare case of wrapping of t, subtraction should fix it.
        int dTicks = (t - last_time) / delay;
        Ticks += dTicks;

        // last_time is now dTicks * delay later
        last_time += dTicks * delay;
    };
}

sdl::timer::timer(int msec)
    : last_time(0), running(false), delay(msec), Ticks(0)
{
}

void sdl::timer::set_delay(int msec)
{
    update();
    delay = msec;
}

int sdl::timer::get_delay()
{
    return delay;
}

void sdl::timer::start()
{
    last_time = SDL_GetTicks();
    running = true;
    Ticks = 0;
}

void sdl::timer::stop()
{
    update();
    running = false;
}

void sdl::timer::wait()
{
    if (running) {
        do {
            // similar to update()
            Uint32 t = SDL_GetTicks();
            // in the rare case of wrapping of t, subtraction should fix it
            Uint32 dtime = t - last_time;
            int dTicks = dtime / delay;
            int rem = dtime % delay;
            Ticks += dTicks;
            last_time += dtime - rem;

            // if no pending ticks, perform a non-busy wait until
            // about next tick
            if (Ticks < 1)
                SDL_Delay(delay - rem);
        } while (Ticks < 1);
    }
}

void sdl::timer::done()
{
    --Ticks;
}

int sdl::timer::ticks()
{
    update();
    return Ticks;
}
