// ship.cc - functions required by ship
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "ball.hh"
#include "controller.hh"

// A move of time t to a ship
void balls::ship::move(int t)
{
    // Modify vel because of collisions.
    v += dv; // dv is change in vel (due to collisions after previous move).
    dv.set(); // reset dv

    // Add velocity * time to distance
    l += v * t;
      
    if (hit_wall(this)) {
        bounce_wall(this);
        get_inside_wall(this);
    }
      
    if (strength <= 0)
        rem = true;
}

// what to do after a collision. Note that location and velocity
// should NOT be changed, as they might be used by other balls before
// the next move.
void balls::ship::do_collide(const ball* other)
{
    if (is_pickup(other)) {
    }
    else {
        int d = other -> damage();
        if (hardness < 1000)
            strength -= d * (1000 - hardness) / 1000;
        hardness -= d;
        if (hardness < 0)
            hardness = 0;

        // for one collision, add to dv, not assign, in case we have to cater
        // for more than one collision in a single move
        // -- Next line commented to avoid ship moving on collision --
        //      dv += dvel_collision(this, other);
    }
}

// damage caused by a ship
int balls::ship::damage() const
{
    return mas;
}

balls::the_ship::the_ship(vect3d<> ll, vect3d<> vv, draw_ball_info* inf,
                          control::controller* controller)
    : ship((*inf->bd)["the_ship"]["radius"],
           (*inf->bd)["the_ship"]["mass"],
           ll, vv, inf,
           (*inf->bd)["the_ship"]["strength"],
           (*inf->bd)["the_ship"]["hardness"],
           (*inf->bd)["the_ship"]["ammo"]),
      graphics::deep_drawable((*inf->gallery)["the_ship"],
                              inf->mapp),
      min_zvel((*inf->bd)["the_ship"]["minvelz"]),
      max_zvel((*inf->bd)["the_ship"]["maxvelz"]),
      zaccel((*inf->bd)["the_ship"]["accelz"]),
      zdecel((*inf->bd)["the_ship"]["decelz"]),
      xyvel((*inf->bd)["the_ship"]["velxy"]),
      depth_scr((*inf->bd)["the_ship"]["depth_scr"]),
      depth_scr_till_now(0),
      control(controller),
      changed_weap_last_turn(false)
{
    weapons.push_back(new def_weapon(this));
    weapons.push_back(new fast_weapon(this));
    cur_weap = weapons.begin();
}

balls::the_ship::~the_ship()
{
}

// move the ship
void balls::the_ship::move(int t)
{
    v.setx(0.0).sety(0.0);

    // get the action
    control->poll();

    bool left = control->left(), right = control->right();
    bool up = control->up(), down = control->down();
    bool accel = control->fire4(), decel = control->fire2();
    bool fire = control->fire3(), weapon = control->fire1();

    // move left
    if (left && !right)
        v.setx(-xyvel);
    // move right
    if (right && !left)
        v.setx(xyvel);
    // move up
    if (up && !down)
        v.sety(-xyvel);
    // move down
    if (down && !up)
        v.sety(xyvel);

    // accelerate
    if (accel && !decel)
        v += vectz<>(zaccel * t);
    // decelerate
    if (decel && !accel)
        v -= vectz<>(zdecel * t);
   
    if (v.z() < min_zvel)
        v.setz(min_zvel);
    else if (v.z() > max_zvel)
        v.setz(max_zvel);
   
    l += v * t;
    
    if (hit_wall(this))
        get_inside_wall(this);
    if (strength <= 0)
        rem = true;

    // toggle weapon
    if (changed_weap_last_turn) {
        if (!weapon)
            changed_weap_last_turn = false;
    }
    else {
        if (weapon) {
            changed_weap_last_turn = true;
            // if previous weapon is valid, move to next weapon
            if (cur_weap != weapons.end())
                ++cur_weap;
            // if previous weapon is invalid or we have reached end of list,
            // move to beginning of list. Note that if there are no weapons,
            // cur_weap == weapons->end() both before and after.
            if (cur_weap == weapons.end())
                cur_weap = weapons.begin();
        }
    }

    // fire
    if (cur_weap != weapons.end()) {
        if (fire)
            (*cur_weap)->fire();
        (*cur_weap)->charge(t);
    }
    
    // add score for depth
    int req_score = int(l.z() * depth_scr);
    add_score(req_score - depth_scr_till_now);
    depth_scr_till_now = req_score;
}

balls::def_weapon::def_weapon(ship* sp)
    : weapon(sp),
      ch(0),
      req_charge((*sp->inf()->bd)["def_weapon"]["req_charge"]),
      req_ammo((*sp->inf()->bd)["def_weapon"]["req_ammo"]),
      extra_vel((*sp->inf()->bd)["def_weapon"]["extra_vel"]),
      im((*static_cast<draw_ball_info*>(sp->inf())->wicons)["def_weapon"])
{
}

void balls::def_weapon::charge(int t)
{
    ch += t;
}

void balls::def_weapon::fire()
{
    if (ch < req_charge || ammo() < req_ammo)
        return;
    ch = 0;
    ammo() -= req_ammo;
    vect3d<> bloc(s->loc()), bvel(vectz<>(s->vel().z()));
    bloc += vectz<>(s->radius() * 3);
    bvel += vectz<>(extra_vel);
    ball* bullet = new normal_bullet(bloc, bvel,
                                     static_cast<draw_ball_info*>(s->inf()),
                                     s);
    s->inf()->swal -> swallow(bullet);
}

balls::fast_weapon::fast_weapon(ship* sp)
    : weapon(sp),
      ch(0),
      req_charge((*sp->inf()->bd)["fast_weapon"]["req_charge"]),
      req_ammo((*sp->inf()->bd)["fast_weapon"]["req_ammo"]),
      extra_vel((*sp->inf()->bd)["fast_weapon"]["extra_vel"]),
      im((*static_cast<draw_ball_info*>(sp->inf())->wicons)["fast_weapon"])
{
}

void balls::fast_weapon::charge(int t)
{
    ch += t;
}

void balls::fast_weapon::fire()
{
    if (ch < req_charge || ammo() < req_ammo)
        return;
    ch = 0;
    ammo() -= req_ammo;
    vect3d<> bloc(s->loc()), bvel(vectz<>(s->vel().z()));
    bloc += vectz<>(s->radius() * 3);
    bvel += vectz<>(extra_vel);
    ball* bullet = new normal_bullet(bloc, bvel,
                                     static_cast<draw_ball_info*>(s->inf()),
                                     s);
    s->inf()->swal -> swallow(bullet);
}
